"""Abstract base class for stream evaluators."""

from portia import Config

from steelthread.streams.metrics import StreamMetric
from steelthread.streams.models import PlanRunStreamItem, PlanStreamItem


class StreamEvaluator:
    """Abstract base class for implementing stream evaluation logic.

    Subclasses must define logic to evaluate either a plan or a plan run,
    typically sourced from pre-recorded executions (e.g. production runs).

    Attributes:
        config (Config): Portia configuration object.

    """

    def __init__(self, config: Config) -> None:
        """Initialize the evaluator with a Portia config.

        Args:
            config (Config): Portia configuration containing model info and credentials.

        """
        super().__init__()
        self.config = config

    def process_plan(
        self,
        stream_item: PlanStreamItem,  # noqa: ARG002
    ) -> list[StreamMetric] | StreamMetric | None:
        """Process a Plan stream item.

        Args:
            stream_item (PlanStreamItem): The Plan to evaluate.

        Returns:
            list[StreamMetric] | StreamMetric: Metric(s) resulting from evaluation.

        """
        return []

    def process_plan_run(
        self,
        stream_item: PlanRunStreamItem,  # noqa: ARG002
    ) -> list[StreamMetric] | StreamMetric | None:
        """Process a PlanRunStream item.

        Args:
            stream_item (PlanRunStreamItem): The item to evaluate

        Returns:
            list[StreamMetric] | StreamMetric | None: Metric(s) or None if not applicable.

        """
        return []



"""Backend for Portia evals."""

import httpx
from portia import (
    Config,
    LocalDataValue,
    Plan,
    PlanRun,
    PlanRunState,
    PlanUUID,
)
from portia.plan_run import PlanRunOutputs, PlanRunUUID
from portia.storage import PortiaCloudClient
from pydantic import BaseModel

from steelthread.streams.models import (
    PlanRunStreamItem,
    PlanStreamItem,
    Stream,
)


class PortiaStreamBackend(BaseModel):
    """Client interface for interacting with the Portia API for evaluations.

    Provides methods to load test cases and mark them as processed.

    Attributes:
        config (Config): The Portia configuration containing API credentials and context.

    """

    config: Config

    def client(self) -> httpx.Client:
        """Create an HTTP client for interacting with the Portia API.

        Returns:
            httpx.Client: A configured HTTP client.

        """
        return PortiaCloudClient().get_client(self.config)

    def check_response(self, response: httpx.Response) -> None:
        """Validate the response from Portia API.

        Args:
            response (httpx.Response): The response from the Portia API to check.

        Raises:
            ValueError: If the response status code indicates an error.

        """
        if not response.is_success:
            error_str = str(response.content)
            raise ValueError(error_str)

    def get_stream(self, stream_name: str) -> Stream:
        """Load information about a stream.

        Args:
            stream_name (str): The name of the stream

        Returns:
            Stream: Information on the stream

        """
        client = self.client()
        url = f"/api/v0/evals/streams/by-name/{stream_name}/"
        response = client.get(url)
        self.check_response(response)
        return Stream(**response.json())

    def load_plan_stream_items(self, stream_id: str, batch_size: int) -> list[PlanStreamItem]:
        """Load stream items from the Portia API with pagination."""
        client = self.client()
        url = f"/api/v0/evals/stream-items/?stream_id={stream_id}"
        test_cases = []

        while url and len(test_cases) < batch_size:
            response = client.get(url)
            self.check_response(response)
            data = response.json()
            for tc in data.get("results", []):
                while len(test_cases) < batch_size:
                    test_cases.append(
                        PlanStreamItem(
                            stream=stream_id,
                            stream_item=tc["id"],
                            plan=Plan(**tc["plan"]),
                        )
                    )
            url = data.get("next")

        return test_cases

    def load_plan_run_stream_items(
        self, stream_id: str, batch_size: int
    ) -> list[PlanRunStreamItem]:
        """Load stream items from the Portia API with pagination."""
        client = self.client()
        url = f"/api/v0/evals/stream-items/?stream_id={stream_id}"
        test_cases = []
        while url and len(test_cases) < batch_size:
            response = client.get(url)
            self.check_response(response)
            data = response.json()
            for tc in data.get("results", []):
                while len(test_cases) < batch_size:
                    test_cases.append(
                        PlanRunStreamItem(
                            stream=stream_id,
                            stream_item=tc["id"],
                            plan=Plan.from_response(tc["plan"]),
                            plan_run=PlanRun(
                                id=PlanRunUUID.from_string(tc["plan_run"]["id"]),
                                plan_id=PlanUUID.from_string(tc["plan_run"]["plan"]["id"]),
                                end_user_id=tc["plan_run"]["end_user"],
                                current_step_index=tc["plan_run"]["current_step_index"],
                                state=PlanRunState(tc["plan_run"]["state"]),
                                outputs=PlanRunOutputs.model_validate(tc["plan_run"]["outputs"]),
                                plan_run_inputs={
                                    key: LocalDataValue.model_validate(value)
                                    for key, value in tc["plan_run"]["plan_run_inputs"].items()
                                },
                            ),
                        )
                    )
            url = data.get("next")

        return test_cases

    def mark_processed(self, item: PlanStreamItem | PlanRunStreamItem) -> None:
        """Mark a stream item as processed in the Portia API.

        Args:
            item (StreamItem): The stream item to mark as processed.

        """
        client = self.client()
        response = client.patch(
            url="/api/v0/evals/stream-items/",
            json={"processed": True, "id": str(item.stream_item)},
        )
        self.check_response(response)



"""Stream Models."""

import enum
from typing import Any

from portia import Plan, PlanRun
from pydantic import BaseModel


class StreamSource(enum.Enum):
    """The source of a stream."""

    PLAN = "plan"
    PLAN_RUN = "plan_run"


class Stream(BaseModel):
    """Definition of a Stream.

    Represents a stream item containing both a plan only.

    Attributes:
         id (str): Unique identifier for the test case.
         plan (Plan): The plan from the StreamItem
         plan_run (PlanRun | None): The plan_run from the StreamItem

    """

    id: str
    name: str
    source: StreamSource
    sample_rate: int
    sample_filters: dict[str, Any]
    last_sampled: str


class PlanStreamItem(BaseModel):
    """Definition of a StreamItem.

    Represents a stream item containing both a plan only.

    Attributes:
         id (str): Unique identifier for the test case.
         plan (Plan): The plan from the StreamItem
         plan_run (PlanRun | None): The plan_run from the StreamItem

    """

    stream: str
    stream_item: str
    plan: Plan


class PlanRunStreamItem(BaseModel):
    """Definition of a PlanRunStreamItem.

    Represents a stream item containing both a plan + plan_run.

    Attributes:
        id (str): Unique identifier for the test case.
        plan (Plan): The plan from the StreamItem
        plan_run (PlanRun): The plan_run from the StreamItem

    """

    stream: str
    stream_item: str
    plan: Plan
    plan_run: PlanRun



"""Tool stubs + registry."""

from collections.abc import Callable
from typing import Any

from portia import (
    Clarification,
    Tool,
    ToolRegistry,
    ToolRunContext,
)
from portia.common import combine_args_kwargs
from portia.tool_call import ToolCallRecord, ToolCallStatus
from pydantic import BaseModel, Field


class ToolStubContext(BaseModel):
    """Context passed to tool stubs."""

    tool_call_index: int
    original_context: ToolRunContext
    original_tool: Tool | None
    args: tuple[Any, ...]
    kwargs: dict[str, Any]


ToolResponseStub = Callable[[ToolStubContext], Any]


class ToolStub(Tool):
    """A tool stub that returns pre-canned data and records tool calls.

    This is useful for testing or evaluating agentic workflows without invoking real tools.

    Attributes:
        child_tool (Tool | None): If set, the stub will defer to this tool for execution.
        return_callable (ToolResponseStub | None): A function used to return fake tool outputs.
        tool_calls (list[ToolCallRecord]): A record of all tool calls made to this stub.

    """

    child_tool: Tool | None = Field(
        default=None,
        description="The child tool. If set output will be fetched from this tool.",
    )
    return_callable: ToolResponseStub | None = Field(
        default=None,
        description="A callable to produce tool outputs. "
        "Takes (call_index, ctx, args, kwargs) as arguments.",
    )
    tool_calls: list[ToolCallRecord] = Field(
        description="A list of all the tool calls this tool has seen."
    )

    def run(
        self,
        ctx: ToolRunContext,
        *args: Any,
        **kwargs: Any,
    ) -> Any:  # noqa: ANN401
        """Run the stub tool and record the call.

        If `child_tool` is provided, it delegates the call.
        If `return_callable` is provided, it uses it to simulate output.
        Otherwise, it raises a RuntimeError.

        Args:
            ctx (ToolRunContext): The tool run context.
            *args (Any): Positional arguments to the tool.
            **kwargs (Any): Keyword arguments to the tool.

        Returns:
            Any: The tool output (either simulated or delegated).

        """
        call_index = len(self.tool_calls)
        tool_call_status = ToolCallStatus.SUCCESS

        if self.return_callable:
            try:
                stub_ctx = ToolStubContext(
                    tool_call_index=call_index,
                    original_context=ctx,
                    args=args,
                    kwargs=kwargs,
                    original_tool=self.child_tool,
                )
                tool_output = self.return_callable(stub_ctx)
            except Exception as e:  # noqa: BLE001
                tool_output = str(e)
                tool_call_status = ToolCallStatus.FAILED
        elif self.child_tool:
            try:
                tool_output = self.child_tool.run(ctx, *args, **kwargs)
            except Exception as e:  # noqa: BLE001
                tool_output = str(e)
                tool_call_status = ToolCallStatus.FAILED
        else:
            raise RuntimeError("ToolStub must have either child_tool or return_callable set.")

        if isinstance(tool_output, Clarification):
            tool_output.plan_run_id = ctx.plan_run.id

        tc = ToolCallRecord(
            tool_name=self.id,
            plan_run_id=ctx.plan_run.id,
            step=ctx.plan_run.current_step_index,
            end_user_id=ctx.end_user.external_id,
            status=tool_call_status,
            input=combine_args_kwargs(*args, **kwargs),
            output=tool_output,
            latency_seconds=0,
        )
        self.tool_calls.append(tc)
        return tool_output


class ToolStubRegistry(ToolRegistry):
    """A registry that allows setting tool stubs while preserving regular tool behavior.

    This is useful for testing: it wraps a real ToolRegistry but replaces some tools
    with stubbed versions that simulate behavior.

    Attributes:
        stubs (dict[str, ToolResponseStub]): A mapping of tool IDs to response stubs.
        stubbed_tools (dict[str, ToolStub]): Cached stubbed tool instances.

    """

    def __init__(self, registry: ToolRegistry, stubs: dict[str, ToolResponseStub]) -> None:
        """Initialize the stub registry.

        Args:
            registry (ToolRegistry): The original registry to wrap.
            stubs (dict[str, ToolResponseStub]): Stub response functions keyed by tool ID.

        """
        super().__init__(registry.get_tools())
        self.stubs = stubs
        self.stubbed_tools: dict[str, ToolStub] = {}

    def get_tool_calls(self, tool_id: str | None = None) -> list[ToolCallRecord]:
        """Get recorded tool calls for a specific stubbed tool or all stubbed tools.

        Args:
            tool_id (str | None): The tool ID to filter by, or None for all.

        Returns:
            list[ToolCallRecord]: A list of tool call records.

        """
        if not tool_id:
            all_calls = []
            for tool_details in self.stubbed_tools.values():
                all_calls.extend(tool_details.tool_calls)
            return all_calls
        if tool_id in self.stubbed_tools:
            return self.stubbed_tools[tool_id].tool_calls
        return []

    def get_tool(self, tool_id: str) -> Tool:
        """Get a stubbed or wrapped tool by ID.

        If the tool has a response stub, a ToolStub is returned that uses it.
        If not, a ToolStub is returned that wraps the original tool.

        Args:
            tool_id (str): The ID of the tool.

        Returns:
            Tool: A stubbed or wrapped tool instance.

        """
        tool = super().get_tool(tool_id)
        if tool.id in self.stubbed_tools:
            return self.stubbed_tools[tool.id]

        if tool_id in self.stubs:
            tool_stub = ToolStub(
                id=tool.id,
                name=tool.name,
                description=tool.description,
                args_schema=tool.args_schema,
                output_schema=tool.output_schema,
                should_summarize=tool.should_summarize,
                return_callable=self.stubs[tool.id],
                tool_calls=[],
            )
        else:
            tool_stub = ToolStub(
                id=tool.id,
                name=tool.name,
                description=tool.description,
                args_schema=tool.args_schema,
                output_schema=tool.output_schema,
                should_summarize=tool.should_summarize,
                child_tool=tool,
                return_callable=None,
                tool_calls=[],
            )

        self.stubbed_tools[tool_id] = tool_stub
        return tool_stub

    def get_tools(self) -> list[Tool]:
        """Get all tools from the registry, replacing them with stubs as needed.

        Returns:
            list[Tool]: A list of stubbed tool instances.

        """
        tools = super().get_tools()
        return [self.get_tool(tool.id) for tool in tools]



"""Core evaluator abstraction."""

from abc import ABC, abstractmethod

from portia import Config, Plan, PlanRun
from portia.storage import ToolCallRecord
from pydantic import BaseModel

from steelthread.evals.metrics import EvalMetric
from steelthread.evals.models import EvalTestCase


class PlanRunMetadata(BaseModel):
    """Model that records metadata for a plan run.

    Attributes:
        tool_calls (list[ToolCallRecord]): A list of tool calls made during the run.
        latency_ms (float): Latency in milliseconds for the plan run.

    """

    tool_calls: list[ToolCallRecord]
    latency_ms: float


class Evaluator(ABC):
    """Abstract base class for implementing evaluation logic.

    Subclasses should implement the `eval_test_case` method to evaluate
    a `PlanRun` against an `EvalTestCase` and return one or more EvalMetrics.

    Attributes:
        config (Config): Portia configuration instance for access to model or tooling info.

    """

    def __init__(self, config: Config) -> None:
        """Initialize the evaluator with a Portia config.

        Args:
            config (Config): Configuration object for Portia and LLM integration.

        """
        super().__init__()
        self.config = config

    @abstractmethod
    def eval_test_case(
        self,
        test_case: EvalTestCase,
        final_plan: Plan,
        final_plan_run: PlanRun,
        additional_data: PlanRunMetadata,
    ) -> list[EvalMetric] | EvalMetric | None:
        """Evaluate a test case given its plan run result and metadata.

        Args:
            test_case (EvalTestCase): The test case defining expected behavior/assertions.
            final_plan (Plan): The plan to evaluate.
            final_plan_run (PlanRun): The plan run output to evaluate.
            additional_data (PlanRunMetadata): Metadata like latency and tool call history.

        Returns:
            list[EvalMetric] | EvalMetric | None: One or more EvalMetrics representing results.

        """
        return []



"""Main runner for steel thread."""

from portia import Portia

from steelthread.evals import EvalConfig, EvalRunner
from steelthread.streams import StreamConfig, StreamProcessor


class SteelThread:
    """Main steel thread runner.

    Provides static methods to run both stream and evaluation based workflows.
    """

    @staticmethod
    def process_stream(config: StreamConfig) -> None:
        """Process stream items based on the config given.

        Args:
            config (EvalConfig): Configuration for stream processor.

        """
        StreamProcessor(config).run()

    @staticmethod
    def run_evals(portia: Portia, config: EvalConfig) -> None:
        """Run evaluations using Portia and the provided configuration.

        Args:
            portia (Portia): Portia instance used for model access and execution.
            config (EvalConfig): Configuration for evaluation runs.

        """
        EvalRunner(portia, config).run()



"""LLM as Judge implementation."""

from portia import Config, Output, Plan
from portia.plan_run import PlanRun

from steelthread.evals.evaluator import Evaluator, PlanRunMetadata
from steelthread.evals.metrics import EvalMetric
from steelthread.evals.models import (
    Assertion,
    EvalTestCase,
    FinalOutputAssertion,
    LatencyAssertion,
    LLMAsJudgeAssertion,
    OutcomeAssertion,
    ToolCallsAssertion,
)
from steelthread.utils.llm import LLMScorer, MetricOnly


class OutputScoreCalculator:
    """Calculate the output score using simple string matching logic."""

    @staticmethod
    def calculate(output: Output | None, assertion: FinalOutputAssertion) -> float:
        """Calculate a score comparing the output to an expected value.

        Args:
            output (Output | None): The actual output from the plan run.
            assertion (FinalOutputAssertion): The expected output assertion.

        Returns:
            float: 1.0 for match, 0.0 otherwise.

        """
        output_str = str(output.get_value()) if output else ""
        expected_str = str(assertion.value)

        match assertion.output_type:
            case "exact_match":
                return 1.0 if output_str == expected_str else 0.0
            case "partial_match":
                return 1.0 if expected_str in output_str else 0.0
            case _:
                raise ValueError(f"Unknown output_type: {assertion.output_type}")


class AssertionEvaluator:
    """Evaluate assertions defined in test cases against a PlanRun."""

    def __init__(
        self,
        config: Config,
        test_case: EvalTestCase,
        plan: Plan,
        plan_run: PlanRun,
        metadata: PlanRunMetadata,
    ) -> None:
        """Initialize the evaluator with Portia config and run data.

        Args:
            config (Config): Portia config for model access or scoring.
            test_case (EvalTestCase): The test case.
            plan (Plan): The linked plan.
            plan_run (PlanRun): The plan run to evaluate.
            metadata (PlanRunMetadata): Additional data about the run (e.g., latency, tool calls).

        """
        self.config = config
        self.test_case = test_case
        self.plan = plan
        self.plan_run = plan_run
        self.metadata = metadata

    def evaluate(self, assertion: Assertion) -> list[EvalMetric]:
        """Evaluate a single assertion and return one or more EvalMetrics.

        Args:
            assertion (Assertion): The assertion to evaluate.

        Returns:
            list[EvalMetric]: One or more EvalMetric results.

        """
        match assertion.type:
            case "outcome":
                return [self._evaluate_outcome(assertion)]
            case "final_output":
                return self._evaluate_final_output(assertion)
            case "latency":
                return [self._evaluate_latency(assertion)]
            case "tool_calls":
                return [self._evaluate_tool_calls(assertion)]
            case "llm_as_judge":
                return self._evaluate_llm_judge(assertion)
            case "custom":
                return []
            case _:
                raise ValueError(f"Unsupported assertion type: {assertion.type}")

    def _evaluate_llm_judge(self, assertion: LLMAsJudgeAssertion) -> list[EvalMetric]:
        scorer = LLMScorer(self.config)
        metrics = scorer.score(
            task_data=[
                f"Please score the given plan run based on these rules. Rules:{assertion.value}",
                self.plan_run.model_dump_json(),
            ],
            metrics_to_score=[
                MetricOnly(
                    name=assertion.type,
                    description="LLM-based score",
                )
            ],
        )
        return [
            EvalMetric.from_test_case(
                test_case=self.test_case,
                score=m.score,
                name=m.name,
                expectation=assertion.value,
                description=m.description,
                explanation=m.explanation,
            )
            for m in metrics
        ]

    def _evaluate_outcome(self, assertion: OutcomeAssertion) -> EvalMetric:
        """Evaluate the final state of the plan run."""
        assertion_value = assertion.value.lower()
        actual_value = self.plan_run.state.lower()

        score = 1 if assertion_value == actual_value else 0
        return EvalMetric.from_test_case(
            test_case=self.test_case,
            score=score,
            name=assertion.type,
            expectation=assertion_value,
            actual_value=actual_value,
            description="Whether the outcome exactly matches expected value",
        )

    def _evaluate_final_output(self, assertion: FinalOutputAssertion) -> list[EvalMetric]:
        """Evaluate the final output using either string comparison or LLM-based scoring."""
        assertion_value = assertion.value
        actual_value = str(
            self.plan_run.outputs.final_output.get_value()
            if self.plan_run.outputs.final_output
            else ""
        )

        if assertion.output_type == "llm_judge":
            scorer = LLMScorer(self.config)
            metrics = scorer.score(
                task_data=[
                    f"Please score based on how well the output matches {assertion.value}",
                    self.plan_run.model_dump_json(),
                ],
                metrics_to_score=[
                    MetricOnly(
                        name=assertion.type,
                        description="LLM-based final output score",
                    )
                ],
            )
            return [
                EvalMetric.from_test_case(
                    test_case=self.test_case,
                    score=m.score,
                    name=m.name,
                    expectation=assertion_value,
                    actual_value=actual_value,
                    description=m.description,
                    explanation=m.explanation,
                )
                for m in metrics
            ]

        score = OutputScoreCalculator.calculate(self.plan_run.outputs.final_output, assertion)
        return [
            EvalMetric.from_test_case(
                test_case=self.test_case,
                score=score,
                name=assertion.type,
                expectation=assertion_value,
                actual_value=actual_value,
                description="Exact or partial final output match",
            )
        ]

    def _evaluate_latency(self, assertion: LatencyAssertion) -> EvalMetric:
        """Evaluate the latency against a threshold using normalized difference."""
        actual = self.metadata.latency_ms
        target = assertion.threshold_ms
        score = 1 - (abs(target - actual) / max(abs(target), abs(actual), 1e-8))
        return EvalMetric.from_test_case(
            test_case=self.test_case,
            score=score,
            name=assertion.type,
            expectation=str(target),
            actual_value=str(actual),
            description="Normalized latency score",
        )

    def _evaluate_tool_calls(self, assertion: ToolCallsAssertion) -> EvalMetric:
        """Evaluate whether expected tools were called (or not called)."""
        expected_calls = 0
        actual_calls = 0

        for tool_call_name, expectation in assertion.calls.items():
            matched = [tc for tc in self.metadata.tool_calls if tc.tool_name == tool_call_name]
            if expectation.called:
                expected_calls += 1
                actual_calls += 1 if matched else 0
            elif matched:
                actual_calls += 1

        if expected_calls == 0 and actual_calls == 0:
            score = 1.0
        elif expected_calls > 0:
            score = min(actual_calls / expected_calls, 1.0)
        else:
            score = 0.0

        return EvalMetric.from_test_case(
            test_case=self.test_case,
            score=score,
            name=assertion.type,
            expectation=[
                tool_name for tool_name in assertion.calls if assertion.calls[tool_name].called
            ],
            actual_value=[tc.tool_name for tc in self.metadata.tool_calls],
            description="Tool call usage score",
        )


class DefaultEvaluator(Evaluator):
    """Default implementation of an evaluator that evaluates test case assertions."""

    def eval_test_case(
        self,
        test_case: EvalTestCase,
        final_plan: Plan,
        final_plan_run: PlanRun,
        additional_data: PlanRunMetadata,
    ) -> list[EvalMetric] | None:
        """Evaluate all assertions defined in the test case.

        Args:
            test_case (TestCase): The test case to evaluate.
            final_plan (Plan): The executed plan to evaluate.
            final_plan_run (PlanRun): The executed plan run to evaluate.
            additional_data (PlanRunMetadata): Additional context like latency, tool usage.

        Returns:
            list[EvalMetric] | None: A list of EvalMetrics derived from assertions, or None if none.

        """
        evaluator = AssertionEvaluator(
            self.config, test_case, final_plan, final_plan_run, additional_data
        )
        all_metrics = []
        for assertion in test_case.assertions:
            all_metrics.extend(evaluator.evaluate(assertion))
        return all_metrics



"""Stream Metrics."""

from abc import ABC, abstractmethod

import httpx
import pandas as pd
from portia import Config
from portia.storage import PortiaCloudClient
from pydantic import BaseModel, Field, field_validator

from steelthread.streams.models import PlanRunStreamItem, PlanStreamItem

MIN_EXPLANATION_LENGTH = 10


class StreamMetric(BaseModel):
    """A single record of an observation.

    Attributes:
        stream (str): The id of the stream this metric relates to
        stream_item (str): The id of the stream item this metric relates to
        score (float): The numeric value of the metric.
        name (str): The name of the metric.
        description (str): A human-readable description of the metric.
        explanation (str | None): An optional explanation of the score.
        tags (dict[str, str]): A set of tags to query this metric by.

    """

    stream: str
    stream_item: str
    score: float
    name: str
    description: str
    explanation: str | None = Field(
        default=None,
        description="An optional explanation of the score.",
    )
    tags: dict[str, str] | None = Field(default={}, description="Tags for querying this metric.")

    @classmethod
    def from_stream_item(
        cls,
        stream_item: PlanStreamItem | PlanRunStreamItem,
        score: float,
        name: str,
        description: str,
        explanation: str | None = None,
    ) -> "StreamMetric":
        """Create a metric from a stream item.

        Args:
            stream_item (PlanStreamItem | PlanRunStreamItem): The stream item this metric relates to
            score (float): The numeric value of the metric.
            name (str): The name of the metric.
            description (str): A human-readable description of the metric.
            explanation (str | None): An optional explanation of the score.

        """
        return cls(
            stream=stream_item.stream,
            stream_item=stream_item.stream_item,
            score=score,
            name=name,
            description=description,
            explanation=explanation,
        )

    @field_validator("explanation")
    @classmethod
    def explanation_min_length(cls, v: str | None) -> str | None:
        """If an explanation is provided it must have length."""
        if v is not None and len(v) < MIN_EXPLANATION_LENGTH:
            raise ValueError("explanation must be at least 5 characters long")
        return v


class StreamMetricsBackend(ABC):
    """Abstract interface for saving metrics."""

    @abstractmethod
    def save_metrics(self, metrics: list[StreamMetric]) -> None:
        """Save a list of tagged metrics for a specific evaluation run.

        Args:
            metrics (list[StreamMetricWithTags]): The metrics to save.

        """
        raise NotImplementedError


class PortiaStreamMetricsBackend(StreamMetricsBackend):
    """Backend for saving stream metrics to the Portia API."""

    def __init__(self, config: Config) -> None:
        """Init config."""
        self.config = config

    def client(self) -> httpx.Client:
        """Return an authenticated HTTP client."""
        return PortiaCloudClient().get_client(self.config)

    def check_response(self, response: httpx.Response) -> None:
        """Raise if response is not successful."""
        if not response.is_success:
            raise ValueError(f"Portia API error: {response.status_code} - {response.text}")

    def save_metrics(self, metrics: list[StreamMetric]) -> None:
        """Send metrics to the Portia API for a given eval run."""
        payload = [m.model_dump() for m in metrics]

        client = self.client()
        response = client.post("/api/v0/evals/stream-metrics/", json=payload)
        self.check_response(response)


class StreamLogMetricBackend(StreamMetricsBackend):
    """Implementation of the metrics backend that logs scores.

    This backend prints average metric scores grouped by name and tags.
    """

    def save_metrics(self, metrics: list[StreamMetric]) -> None:
        """Log metrics via pandas.

        Converts the metrics list into a DataFrame, expands tags into columns,
        groups by stream_item and name, and prints average scores.

        Args:
            metrics (list[StreamMetric]): The metrics to log.

        """
        flattened = [m.model_dump() for m in metrics]

        # Convert list of metrics to DataFrame
        dataframe = pd.DataFrame(flattened)

        # Expand the 'tags' column into separate columns
        tags_df = dataframe["tags"].apply(pd.Series)
        dataframe = pd.concat([dataframe.drop(columns=["tags"]), tags_df], axis=1)

        # Group by stream_item and metric name
        avg_scores = dataframe.groupby(["stream_item", "name"])["score"].mean().reset_index()

        # Print
        print("\n=== Metric Averages by Stream Item ===")  # noqa: T201
        print(avg_scores.to_string(index=False))  # noqa: T201



"""LLM judge for metrics."""

from portia import Config, Message
from pydantic import BaseModel, Field, field_validator

MIN_EXPLANATION_LENGTH = 10


class MetricOnly(BaseModel):
    """An input to the LLM scorer.

    Attributes:
        name (str): The name of the metric.
        description (str): A human-readable description of the metric.

    """

    name: str
    description: str


class MetricOutput(BaseModel):
    """A single record of an observation.

    Attributes:
        score (float): The numeric value of the metric.
        name (str): The name of the metric.
        description (str): A human-readable description of the metric.

    """

    score: float
    name: str
    description: str
    explanation: str = Field(description="A required explanation of the score.")

    @field_validator("explanation")
    @classmethod
    def explanation_min_length(cls, v: str | None) -> str | None:
        """If an explanation is provided it must have length."""
        if v is not None and len(v) < MIN_EXPLANATION_LENGTH:
            raise ValueError("explanation must be at least 10 characters long")
        return v


class MetricOutputList(BaseModel):
    """A list of metrics."""

    metrics: list[MetricOutput]


class LLMScorer:
    """An implementation of an LLM as Judge to return metrics.

    Uses a configured LLM to score a list of metrics against task data, returning scores
    with optional explanations.
    """

    def __init__(
        self,
        config: Config,
        base_prompt: str = """You are an expert reviewer charged with evaluating agentic executions.
        For each metric provided please provide a score between 0 and 1 based on the data and task
        provided. IMPORTANT - Also include an explanation as to why you score it this way.""",
    ) -> None:
        """Initialize the LLMScorer.

        Args:
            config (Config): Configuration object providing model access.
            base_prompt (str): Instructional prompt used to guide the model.

        """
        self.config = config
        self.base_prompt = base_prompt

    def score(
        self,
        task_data: list[str],
        metrics_to_score: list[MetricOnly],
    ) -> list[MetricOutput]:
        """Scores the given metrics based on the task data.

        Constructs a prompt using the base prompt, metrics, and task data,
        sends it to the model, and parses the structured response.

        Args:
            task_data (list[str]): Input data related to the task being evaluated.
            metrics_to_score (list[Metric]): The metrics to score using the model.

        Returns:
            list[Metric]: The scored metrics.

        """
        messages = [
            Message(role="user", content=self.base_prompt),
            Message(
                role="user",
                content="\n".join(
                    [
                        f"name={metric.name} description={metric.description}"
                        for metric in metrics_to_score
                    ]
                ),
            ),
            Message(role="user", content="\n".join(task_data)),
        ]

        metrics = (
            self.config.get_default_model()
            .get_structured_response(messages, MetricOutputList)
            .metrics
        )
        class_name = self.__class__.__name__
        [
            # use the name of the class here
            print(f"[{class_name}] {metric.name}:{metric.score} explanation: {metric.explanation}")  # noqa: T201
            for metric in metrics
        ]

        return metrics



"""Read only storage."""

from portia import LocalDataValue, Output, Plan, PlanRun, PlanUUID
from portia.plan_run import PlanRunState
from portia.portia import EndUser, InMemoryStorage
from portia.prefixed_uuid import PlanRunUUID
from portia.storage import PlanRunListResponse, Storage, ToolCallRecord


class ReadOnlyStorage(Storage):
    """Gets plans, runs and tool calls from underlying storage, but does not save them.

    Writes are stored only in memory and do not persist to the underlying storage backend.
    Useful for testing, dry runs, or debugging environments.
    """

    def __init__(self, storage: Storage) -> None:
        """Initialize the ReadOnlyStorage instance.

        Args:
            storage (Storage): The underlying storage backend to read from.

        """
        self.storage = storage
        self.local_storage = InMemoryStorage()

    def save_plan(self, plan: Plan) -> None:
        """Save a plan to in-memory storage.

        Args:
            plan (Plan): The plan to save.

        """
        return self.local_storage.save_plan(plan)

    def get_plan(self, plan_id: PlanUUID) -> Plan:
        """Retrieve a plan by ID.

        Args:
            plan_id (PlanUUID): The ID of the plan to fetch.

        Returns:
            Plan: The requested plan.

        """
        try:
            return self.local_storage.get_plan(plan_id)
        except Exception:  # noqa: BLE001
            return self.storage.get_plan(plan_id)

    def get_plan_by_query(self, query: str) -> Plan:
        """Retrieve a plan using a string query.

        Args:
            query (str): Query string to identify a plan.

        Returns:
            Plan: The matching plan.

        """
        try:
            return self.local_storage.get_plan_by_query(query)
        except Exception:  # noqa: BLE001
            return self.storage.get_plan_by_query(query)

    def plan_exists(self, plan_id: PlanUUID) -> bool:
        """Check if a plan exists.

        Args:
            plan_id (PlanUUID): The ID of the plan to check.

        Returns:
            bool: True if the plan exists, False otherwise.

        """
        try:
            return self.local_storage.plan_exists(plan_id)
        except Exception:  # noqa: BLE001
            return self.storage.plan_exists(plan_id)

    def save_plan_run(self, plan_run: PlanRun) -> None:
        """Save a plan run to in-memory storage.

        Args:
            plan_run (PlanRun): The plan run to save.

        """
        return self.local_storage.save_plan_run(plan_run)

    def get_plan_run(self, plan_run_id: PlanRunUUID) -> PlanRun:
        """Retrieve a plan run by ID.

        Args:
            plan_run_id (PlanRunUUID): The ID of the plan run.

        Returns:
            PlanRun: The requested plan run.

        """
        try:
            return self.local_storage.get_plan_run(plan_run_id)
        except Exception:  # noqa: BLE001
            return self.storage.get_plan_run(plan_run_id)

    def get_plan_runs(
        self,
        run_state: PlanRunState | None = None,
        page: int | None = None,
    ) -> PlanRunListResponse:
        """Retrieve a list of plan runs.

        Args:
            run_state (PlanRunState | None): Optional filter by run state.
            page (int | None): Optional pagination.

        Returns:
            PlanRunListResponse: The response containing plan runs.

        """
        try:
            return self.local_storage.get_plan_runs(run_state, page)
        except Exception:  # noqa: BLE001
            return self.storage.get_plan_runs()

    def save_tool_call(self, tool_call: ToolCallRecord) -> None:
        """Save a tool call to in-memory storage.

        Args:
            tool_call (ToolCallRecord): The tool call to record.

        """
        return self.local_storage.save_tool_call(tool_call)

    def save_plan_run_output(
        self,
        output_name: str,
        output: Output,
        plan_run_id: PlanRunUUID,
    ) -> Output:
        """Save an output for a plan run in in-memory storage.

        Args:
            output_name (str): The output key name.
            output (Output): The output value.
            plan_run_id (PlanRunUUID): ID of the related plan run.

        Returns:
            Output: The saved output.

        """
        return self.local_storage.save_plan_run_output(output_name, output, plan_run_id)

    def get_plan_run_output(self, output_name: str, plan_run_id: PlanRunUUID) -> LocalDataValue:
        """Retrieve a plan run output by name and ID.

        Args:
            output_name (str): The output key name.
            plan_run_id (PlanRunUUID): ID of the related plan run.

        Returns:
            LocalDataValue: The retrieved output.

        """
        return self.local_storage.get_plan_run_output(output_name, plan_run_id)

    def get_similar_plans(self, query: str, threshold: float = 0.5, limit: int = 5) -> list[Plan]:
        """Get similar plans from the underlying (read-only) storage.

        Args:
            query (str): Search query to match similar plans.
            threshold (float): Similarity threshold.
            limit (int): Maximum number of results to return.

        Returns:
            list[Plan]: A list of similar plans.

        """
        return self.storage.get_similar_plans(query, threshold, limit)

    def save_end_user(self, end_user: EndUser) -> EndUser:
        """Save an end user in memory.

        Args:
            end_user (EndUser): The end user to save.

        Returns:
            EndUser: The saved user.

        """
        return self.local_storage.save_end_user(end_user)

    def get_end_user(self, external_id: str) -> EndUser | None:
        """Retrieve an end user by external ID.

        Args:
            external_id (str): The external user identifier.

        Returns:
            EndUser | None: The matching user, or None if not found.

        """
        try:
            return self.local_storage.get_end_user(external_id)
        except Exception:  # noqa: BLE001
            return self.storage.get_end_user(external_id)



"""Class for tagging metrics."""

from steelthread.streams.metrics import StreamMetric
from steelthread.streams.models import PlanRunStreamItem, PlanStreamItem


class StreamMetricTagger:
    """Class for attaching tags to metrics."""

    @staticmethod
    def attach_tags(
        metrics: list[StreamMetric] | StreamMetric,
        stream_item: PlanStreamItem | PlanRunStreamItem,  # noqa: ARG004
        additional_tags: dict[str, str] | None = None,
    ) -> list[StreamMetric]:
        """Attach configuration-based and additional tags to a metric.

        Args:
            metrics (list[StreamMetric] | StreamMetric): The original metrics to tag.
            stream_item (PlanStreamItem | PlanRunStreamItem): the item the metric is for.
            additional_tags (dict[str, str] | None): Extra tags to include (optional).

        Returns:
            MetricWithTag: The metric augmented with tags.

        """

        def append_tags(m: StreamMetric) -> StreamMetric:
            m.tags = {
                **(additional_tags or {}),
            }
            return m

        if isinstance(metrics, StreamMetric):
            return [append_tags(metrics)]
        return [append_tags(m) for m in metrics]



"""Contains implementations of streams for SteelThread."""

from .backend import PortiaStreamBackend
from .evaluator import StreamEvaluator
from .llm_as_judge import LLMJudgeEvaluator
from .metrics import (
    PortiaStreamMetricsBackend,
    StreamLogMetricBackend,
    StreamMetric,
    StreamMetricsBackend,
)
from .models import PlanRunStreamItem, PlanStreamItem, Stream, StreamSource
from .stream_processor import StreamConfig, StreamProcessor
from .tags import StreamMetricTagger

__all__ = [
    "LLMJudgeEvaluator",
    "PlanRunStreamItem",
    "PlanStreamItem",
    "PortiaStreamBackend",
    "PortiaStreamMetricsBackend",
    "Stream",
    "StreamConfig",
    "StreamEvaluator",
    "StreamLogMetricBackend",
    "StreamMetric",
    "StreamMetricTagger",
    "StreamMetricsBackend",
    "StreamProcessor",
    "StreamSource",
]



"""Contains implementations of portia specific logic for SteelThread."""

from .portia import NoAuthPullPortia
from .storage import ReadOnlyStorage
from .tools import ToolStub, ToolStubContext, ToolStubRegistry

__all__ = [
    "NoAuthPullPortia",
    "ReadOnlyStorage",
    "ToolStub",
    "ToolStubContext",
    "ToolStubRegistry",
]



"""Stream Processor for steel thread."""

from concurrent.futures import ThreadPoolExecutor, as_completed

from portia import Config
from portia.portia import PortiaCloudStorage

from steelthread.streams.backend import PortiaStreamBackend
from steelthread.streams.evaluator import StreamEvaluator
from steelthread.streams.llm_as_judge import LLMJudgeEvaluator
from steelthread.streams.metrics import (
    PortiaStreamMetricsBackend,
    StreamLogMetricBackend,
    StreamMetric,
    StreamMetricsBackend,
)
from steelthread.streams.models import PlanRunStreamItem, PlanStreamItem, Stream, StreamSource
from steelthread.streams.tags import StreamMetricTagger


class StreamConfig:
    """Configuration for processing streams.

    Attributes:
        stream_name (str): The name of the evals containing test cases.
        portia_config (Config): Configuration for connecting to Portia API.
        iterations (int): Number of times to run each evaluation (default 3).
        evaluators (list[StreamEvaluator]): List of evaluator instances to apply.
        additional_tags (dict[str, str]): Tags to apply to generated metrics.
        metrics_backends (list[MetricsBackend]): Output destinations for metrics.
        max_concurrency (int | None): Maximum number of concurrent tests to run.
        batch_size (int | None): Maximum number of items to process.

    """

    def __init__(
        self,
        stream_name: str,
        config: Config,
        evaluators: list[StreamEvaluator] | None = None,
        additional_tags: dict[str, str] | None = None,
        metrics_backends: list[StreamMetricsBackend] | None = None,
        max_concurrency: int | None = None,
        batch_size: int | None = None,
    ) -> None:
        """Initialize the evaluation configuration.

        Args:
            stream_name (str): stream name to process.
            config (Config): Portia config (must include API key).
            evaluators (list[StreamEvaluator] | None): Evaluators to apply.
            additional_tags (dict[str, str] | None): Extra tags to add to each metric.
            metrics_backends (list[MetricsBackend] | None): Metric writers.
            max_concurrency (int | None): Maximum number of concurrent tests to run.
            batch_size (int | None): Number of items to process.

        """
        config.must_get_api_key("portia_api_key")
        self.stream_name = stream_name
        self.portia_config = config
        self.evaluators = evaluators or [LLMJudgeEvaluator(config)]
        self.additional_tags = additional_tags or {}
        self.metrics_backends = metrics_backends or [
            StreamLogMetricBackend(),
            PortiaStreamMetricsBackend(config),
        ]
        self.max_concurrency = max_concurrency or 5
        self.batch_size = batch_size or 10


class StreamProcessor:
    """Runner for executing stream evaluation test cases and collecting metrics."""

    def __init__(self, config: StreamConfig) -> None:
        """Initialize the runner.

        Args:
            config (StreamConfig): The configuration for the stream.

        """
        self.config = config
        self.backend = PortiaStreamBackend(config=config.portia_config)
        self.storage = PortiaCloudStorage(config.portia_config)

    def run(self) -> None:
        """Execute all test cases in the configured dataset and save metrics.

        - Loads test cases from the backend.
        - Runs each test case for the specified number of iterations.
        - Applies all configured evaluators.
        - Marks cases as processed and writes metrics to backends.
        """
        stream = self.backend.get_stream(self.config.stream_name)

        if stream.source == StreamSource.PLAN:
            return self._process_plan(stream)
        if stream.source == StreamSource.PLAN_RUN:
            return self._process_plan_runs(stream)

        raise ValueError("invalid source")

    def _process_plan(self, stream: Stream) -> None:
        items = self.backend.load_plan_stream_items(
            stream.id,
            self.config.batch_size,
        )
        all_metrics: list[StreamMetric] = []

        futures = []
        with ThreadPoolExecutor(max_workers=self.config.max_concurrency) as executor:
            futures.extend(executor.submit(self._evaluate_plan_stream_item, item) for item in items)

            for future in as_completed(futures):
                result: list[StreamMetric] | StreamMetric | None = future.result()
                if result:
                    all_metrics.extend(result) if isinstance(result, list) else all_metrics.append(
                        result
                    )

        if len(all_metrics) > 0:
            for backend in self.config.metrics_backends:
                backend.save_metrics(all_metrics)

    def _evaluate_plan_stream_item(self, stream_item: PlanStreamItem) -> list[StreamMetric]:
        """Evaluate a single test case across all evaluators."""
        metrics_out = []
        for evaluator in self.config.evaluators:
            metrics = evaluator.process_plan(stream_item)
            if metrics:
                metrics_out.extend(
                    StreamMetricTagger.attach_tags(
                        metrics,
                        stream_item,
                        self.config.additional_tags,
                    )
                )

        self.backend.mark_processed(stream_item)
        return metrics_out

    def _process_plan_runs(self, stream: Stream) -> None:
        items = self.backend.load_plan_run_stream_items(
            stream.id,
            self.config.batch_size,
        )
        all_metrics: list[StreamMetric] = []

        futures = []
        with ThreadPoolExecutor(max_workers=self.config.max_concurrency) as executor:
            futures.extend(
                executor.submit(self._evaluate_plan_run_stream_item, item) for item in items
            )

            for future in as_completed(futures):
                result: list[StreamMetric] | StreamMetric | None = future.result()
                if result:
                    all_metrics.extend(result) if isinstance(result, list) else all_metrics.append(
                        result
                    )

        if len(all_metrics) > 0:
            for backend in self.config.metrics_backends:
                backend.save_metrics(all_metrics)

    def _evaluate_plan_run_stream_item(self, stream_item: PlanRunStreamItem) -> list[StreamMetric]:
        """Evaluate a single test case across all evaluators."""
        metrics_out: list[StreamMetric] = []
        for evaluator in self.config.evaluators:
            metrics = evaluator.process_plan_run(stream_item)
            if metrics:
                metrics_out.extend(
                    StreamMetricTagger.attach_tags(
                        metrics,
                        stream_item,
                        self.config.additional_tags,
                    )
                )

        self.backend.mark_processed(stream_item)
        return metrics_out



"""Metrics backend."""

from abc import ABC, abstractmethod

import httpx
import pandas as pd
from portia.config import Config
from portia.storage import PortiaCloudClient
from pydantic import BaseModel, Field, field_validator

from steelthread.evals.models import EvalTestCase

MIN_EXPLANATION_LENGTH = 10


class EvalMetric(BaseModel):
    """A single record of an observation.

    Attributes:
        dataset (str): the id of the dataset.
        testcase (str): the id of the testcase.
        run (str): the id of the run
        score (float): The numeric value of the metric.
        name (str): The name of the metric.
        expectation (str | list[str] | dict[str, str] | None): expected value
        actual_value (str | list[str] | dict[str, str] | None): actual value
        description (str): A human-readable description of the metric.
        explanation (str | None): An optional explanation of the score.
        tags (dict[str, str]): A set of tags to query this metric by.

    """

    dataset: str
    testcase: str
    run: str
    score: float
    name: str
    description: str
    expectation: str | list[str] | dict[str, str] | None
    actual_value: str | list[str] | dict[str, str] | None
    explanation: str | None = Field(default="", description="An optional explanation of the score.")
    tags: dict[str, str] = Field(default={})

    @field_validator("explanation")
    @classmethod
    def explanation_min_length(cls, v: str | None) -> str | None:
        """If an explanation is provided it must have length."""
        if v is not None and len(v) < MIN_EXPLANATION_LENGTH:
            raise ValueError("explanation must be at least 5 characters long")
        return v

    @classmethod
    def from_test_case(
        cls,
        test_case: EvalTestCase,
        score: float,
        name: str,
        description: str,
        explanation: str | None = None,
        expectation: str | list[str] | dict[str, str] | None = None,
        actual_value: str | list[str] | dict[str, str] | None = None,
    ) -> "EvalMetric":
        """Create a metric from a test case.

        Args:
            test_case (EvalTestCase): The test case this metric relates to
            score (float): The numeric value of the metric.
            name (str): The name of the metric.
            description (str): A human-readable description of the metric.
            explanation (str | None): An optional explanation of the score.
            expectation (str | list[str] | dict[str, str] | None): expected value
            actual_value (str | list[str] | dict[str, str] | None): actual value


        """
        return cls(
            dataset=test_case.dataset,
            testcase=test_case.testcase,
            run=test_case.run,
            score=score,
            name=name,
            description=description,
            explanation=explanation,
            actual_value=actual_value,
            expectation=expectation,
        )


class MetricsBackend(ABC):
    """Abstract interface for saving metrics."""

    @abstractmethod
    def save_eval_metrics(self, metrics: list[EvalMetric]) -> None:
        """Save a list of tagged metrics for a specific evaluation run.

        Args:
            eval_run (EvalRun): The evaluation run context.
            metrics (list[MetricWithTag]): The metrics to save.

        """
        raise NotImplementedError


class PortiaEvalMetricsBackend(MetricsBackend):
    """Backend for saving metrics to the Portia API."""

    def __init__(self, config: Config) -> None:
        """Init config."""
        self.config = config

    def client(self) -> httpx.Client:
        """Return an authenticated HTTP client."""
        return PortiaCloudClient().get_client(self.config)

    def check_response(self, response: httpx.Response) -> None:
        """Raise if response is not successful."""
        if not response.is_success:
            raise ValueError(f"Portia API error: {response.status_code} - {response.text}")

    def save_eval_metrics(self, metrics: list[EvalMetric]) -> None:
        """Send metrics to the Portia API for a given eval run."""
        payload = [m.model_dump() for m in metrics]
        client = self.client()
        response = client.post("/api/v0/evals/eval-metrics/", json=payload)
        self.check_response(response)


class EvalLogMetricBackend(MetricsBackend):
    """Implementation of the metrics backend that logs scores.

    This backend prints average metric scores grouped by name and tags.
    """

    def save_eval_metrics(self, metrics: list[EvalMetric]) -> None:
        """Log metrics via pandas.

        Converts the metrics list into a DataFrame, expands tags into columns,
        groups by metric name and tag combinations, and prints average scores.

        Args:
            eval_run (EvalRun): The evaluation run context (unused).
            metrics (list[MetricWithTag]): The metrics to log.

        """
        # Convert list of metrics to DataFrame
        dataframe = pd.DataFrame([m.model_dump() for m in metrics])

        # Expand the 'tags' column into separate columns
        tags_df = dataframe["tags"].apply(pd.Series)
        dataframe = pd.concat([dataframe.drop(columns=["tags"]), tags_df], axis=1)

        # Determine which columns to group by: metric name + all tag columns
        group_keys = ["name", *tags_df.columns.tolist()]

        # Group by name + tags, then compute mean score
        avg_scores = dataframe.groupby(group_keys)["score"].mean().reset_index()

        # Print
        print("\n=== Metric Averages ===")  # noqa: T201
        print(avg_scores.to_string(index=False))  # noqa: T201



"""Eval runner for steel thread."""

import time
from concurrent.futures import ThreadPoolExecutor, as_completed
from uuid import uuid4

from portia import Config, Plan, PlanRun, Portia
from portia.prefixed_uuid import PlanUUID

from steelthread.evals.backend import PortiaBackend
from steelthread.evals.default_evaluator import DefaultEvaluator
from steelthread.evals.evaluator import Evaluator, PlanRunMetadata
from steelthread.evals.metrics import (
    EvalLogMetricBackend,
    EvalMetric,
    MetricsBackend,
    PortiaEvalMetricsBackend,
)
from steelthread.evals.models import EvalTestCase
from steelthread.evals.tags import EvalMetricTagger
from steelthread.portia.portia import NoAuthPullPortia
from steelthread.portia.storage import ReadOnlyStorage
from steelthread.portia.tools import ToolStubRegistry


class EvalConfig:
    """Configuration for running  evaluations.

    Attributes:
        eval_dataset_name (str): The name of the test eval set to evaluate.
        portia_config (Config): Portia configuration object.
        iterations (int): Number of times each test case should be run (defaults to 3).
        evaluators (list[Evaluator]): List of evaluators to apply to each run.
        additional_tags (dict[str, str]): Tags to attach to each metric result.
        metrics_backends (list[MetricsBackend]): Where to send/save metric results.
        max_concurrency (int | None): Maximum number of concurrent tests to run.

    """

    def __init__(
        self,
        eval_dataset_name: str,
        config: Config,
        iterations: int | None = None,
        evaluators: list[Evaluator] | None = None,
        additional_tags: dict[str, str] | None = None,
        metrics_backends: list[MetricsBackend] | None = None,
        max_concurrency: int | None = None,
    ) -> None:
        """Initialize EvalConfig.

        Args:
            eval_dataset_name (str): Name of the eval set to evaluate.
            config (Config): Portia config with API key.
            iterations (int | None): How many times to run each test case.
            evaluators (list[Evaluator] | None): Evaluators to use (defaults to built-in).
            additional_tags (dict[str, str] | None): Custom tags to attach to metrics.
            metrics_backends (list[MetricsBackend] | None): Output backends (defaults to logger).
            max_concurrency (int | None): Maximum number of concurrent tests to run.

        """
        config.must_get_api_key("portia_api_key")
        self.eval_dataset_name = eval_dataset_name
        self.portia_config = config
        self.iterations = iterations or 3
        self.evaluators = evaluators or [DefaultEvaluator(config)]
        self.additional_tags = additional_tags or {}
        self.metrics_backends = metrics_backends or [
            EvalLogMetricBackend(),
            PortiaEvalMetricsBackend(config),
        ]
        self.max_concurrency = max_concurrency or 5


class EvalRunner:
    """Runner for executing and scoring evaluations."""

    def __init__(self, portia: Portia, config: EvalConfig) -> None:
        """Initialize the runner.

        Wraps the tool registry for stubbing and enforces read-only plan storage.

        Args:
            portia (Portia): Portia engine instance to execute runs.
            config (EvalConfig): Evaluation configuration.

        """
        self.original_portia = portia
        self.config = config
        self.backend = PortiaBackend(config=config.portia_config)

    def _evaluate_and_collect_metrics(self, tc: EvalTestCase) -> list[EvalMetric]:
        """Run a single test case with isolated tool registry and evaluators."""
        inner_registry = self.original_portia.tool_registry
        tool_registry = ToolStubRegistry(inner_registry, stubs={})

        # Patch a local Portia with the test-specific tool registry
        portia = NoAuthPullPortia(config=self.config.portia_config, tools=tool_registry)
        portia.storage = ReadOnlyStorage(portia.storage)  # type: ignore  # noqa: PGH003

        # Run the test case
        plan, plan_run, latency = self._run_test_case(tc, portia)

        # Evaluate with isolated evaluator instances
        all_metrics = []
        for evaluator in self.config.evaluators:
            metrics = evaluator.eval_test_case(
                tc,
                plan,
                plan_run,
                PlanRunMetadata(
                    latency_ms=latency,
                    tool_calls=tool_registry.get_tool_calls(),
                ),
            )
            if metrics:
                all_metrics.extend(
                    EvalMetricTagger.attach_tags_to_test_case(
                        metrics,
                        tc,
                        plan,
                        plan_run,
                        self.config.portia_config,
                        self.config.additional_tags,
                    )
                )
        return all_metrics

    def run(self) -> None:
        """Run the evaluation process.

        - Loads test cases from backend.
        - Executes each test case multiple times.
        - Applies evaluators to generate metrics.
        - Saves metrics using configured backends.

        """
        run_id = str(uuid4())
        test_cases = self.backend.load_evals(self.config.eval_dataset_name, run_id)
        all_metrics = []

        futures = []

        with ThreadPoolExecutor(max_workers=self.config.max_concurrency) as executor:
            futures.extend(
                executor.submit(self._evaluate_and_collect_metrics, tc)
                for tc in test_cases
                for _ in range(self.config.iterations)
            )

            for future in as_completed(futures):
                metrics = future.result()
                if metrics:
                    all_metrics.extend(metrics)

        if len(all_metrics) > 0:
            for backend in self.config.metrics_backends:
                backend.save_eval_metrics(all_metrics)

    def _run_test_case(self, tc: EvalTestCase, portia: Portia) -> tuple[Plan, PlanRun, float]:
        """Execute a single test case and record latency.

        Args:
            tc: The test case to run.
            portia: The instance of portia to use.

        Returns:
            tuple: The plan run output and latency in milliseconds.

        """
        print(f"Executing test case: {tc.input_config.type} - {tc.input_config.value}")  # noqa: T201
        start = time.perf_counter()
        if tc.input_config.type == "query":
            plan = portia.plan(
                tc.input_config.value,
                tools=tc.input_config.tools,
                end_user=tc.input_config.end_user_id,
            )
            output = portia.run_plan(plan)
        elif tc.input_config.type == "plan_id":
            plan = portia.storage.get_plan(PlanUUID.from_string(tc.input_config.value))
            output = portia.run_plan(plan)
        else:
            raise ValueError(f"invalid input_config type: {tc.input_config.type}")
        end = time.perf_counter()
        return plan, output, (end - start) * 1000



"""Class for tagging metrics."""

from portia import Config, Plan, PlanRun

from steelthread.evals.metrics import EvalMetric
from steelthread.evals.models import EvalTestCase


class EvalMetricTagger:
    """Class for attaching tags to metrics."""

    @staticmethod
    def attach_tags_to_test_case(
        metrics: list[EvalMetric] | EvalMetric,
        test_case: EvalTestCase,
        plan: Plan,  # noqa: ARG004
        plan_run: PlanRun | None,  # noqa: ARG004
        config: Config,
        additional_tags: dict[str, str] | None = None,
    ) -> list[EvalMetric]:
        """Attach configuration-based and additional tags to a metric.

        Args:
            metrics (list[EvalMetric] | EvalMetric): the original metrics to tag.
            test_case (EvalTestCase): the original test case.
            plan (Plan): The associated plan
            plan_run (PlanRun): The associated plan_run
            config (Config): Portia config used.
            additional_tags (dict[str, str] | None): Extra tags to include (optional).

        Returns:
            MetricWithTag: The metric augmented with tags.

        """

        def append_tags(m: EvalMetric) -> EvalMetric:
            m.tags = {
                "test_case": str(test_case.testcase),
                "planning_model": config.get_planning_model().model_name,
                "execution_model": config.get_execution_model().model_name,
                "introspection_model": config.get_introspection_model().model_name,
                "summarizer_model": config.get_summarizer_model().model_name,
                **(additional_tags or {}),
            }
            return m

        if isinstance(metrics, EvalMetric):
            return [append_tags(metrics)]
        return [append_tags(m) for m in metrics]



"""Models for test cases."""

from typing import Annotated, Literal

from pydantic import BaseModel, Field


class InputConfig(BaseModel):
    """Configuration for test case input.

    Attributes:
        type (Literal["query", "plan_id"]): The type of input used (query or plan ID).
        value (str): The actual query string or plan ID.
        tools (list[str] | None): Optional list of tools to be used during execution.
        end_user_id (str | None): Optional ID of the end user to simulate context.

    """

    type: Literal["query", "plan_id"]
    value: str
    tools: list[str] | None = None
    end_user_id: str | None = None


class OutcomeAssertion(BaseModel):
    """Assertion for verifying the outcome state of a plan run.

    Attributes:
        type (Literal["outcome"]): Discriminator for the assertion type.
        value (str): Expected plan run state (e.g., "COMPLETE", "FAILED").

    """

    type: Literal["outcome"]
    value: str


class FinalOutputAssertion(BaseModel):
    """Assertion for verifying the final output content.

    Attributes:
        type (Literal["final_output"]): Discriminator for the assertion type.
        output_type (Literal["exact_match", "partial_match", "llm_judge"]): How to eval the output.
        value (str): Expected final output value.

    """

    type: Literal["final_output"]
    output_type: Literal["exact_match", "partial_match", "llm_judge"]
    value: str


class ToolCallAssertion(BaseModel):
    """Assertion record for whether a specific tool was called.

    Attributes:
        called (bool): Whether the tool was expected to be called.

    """

    called: bool


class ToolCallsAssertion(BaseModel):
    """Assertion for verifying calls to tools.

    Attributes:
        type (Literal["tool_calls"]): Discriminator for the assertion type.
        calls (dict[str, ToolCallAssertion]): Mapping of tool names to expected call status.

    """

    type: Literal["tool_calls"]
    calls: dict[str, ToolCallAssertion]


class LatencyAssertion(BaseModel):
    """Assertion for validating runtime latency.

    Attributes:
        type (Literal["latency"]): Discriminator for the assertion type.
        threshold_ms (float): Acceptable maximum latency in milliseconds.

    """

    type: Literal["latency"]
    threshold_ms: float


class LLMAsJudgeAssertion(BaseModel):
    """Assertion for general LLM as judge.

    Attributes:
        type (Literal["llm_as_judge"]): Discriminator for the assertion type.
        value (str): Expected final output value.

    """

    type: Literal["llm_as_judge"]
    value: str


class CustomAssertion(BaseModel):
    """User-defined assertion with arbitrary key-value metadata.

    Attributes:
        type (Literal["custom"]): Discriminator for the assertion type.
        value (dict[str, str]): Arbitrary assertion data.

    """

    type: Literal["custom"]
    value: dict[str, str]


# Discriminated union of all assertion types
Assertion = Annotated[
    OutcomeAssertion
    | FinalOutputAssertion
    | ToolCallsAssertion
    | LatencyAssertion
    | LLMAsJudgeAssertion
    | CustomAssertion,
    Field(discriminator="type"),
]


class EvalTestCase(BaseModel):
    """Model representing an  test case.

    Attributes:
        id (str): Unique identifier for the test case.
        input_config (InputConfig): Configuration for how to run the test.
        assertions (list[Assertion]): Assertions to validate after execution.

    """

    dataset: str
    testcase: str
    run: str
    input_config: InputConfig
    assertions: list[Assertion]

    def get_custom_assertion(self, key: str) -> str | None:
        """Return the value of a custom assertion by key, if it exists.

        Args:
            key (str): Key to search for in the custom assertions.

        Returns:
            str | None: The value if found, otherwise None.

        """
        for assertion in self.assertions:
            if assertion.type == "custom" and key in assertion.value:
                return assertion.value[key]
        return None



"""LLM as Judge implementation."""

from portia import Config

from steelthread.streams.evaluator import StreamEvaluator
from steelthread.streams.metrics import StreamMetric
from steelthread.streams.models import PlanRunStreamItem, PlanStreamItem
from steelthread.utils.llm import LLMScorer, MetricOnly


class LLMJudgeEvaluator(StreamEvaluator):
    """Evaluator that uses an LLM to score Plans and PlanRuns.

    This evaluator uses an LLM-as-Judge approach to assign scores to logical
    properties such as correctness, completeness, and success, based on the
    JSON-serialized plan or run.
    """

    def __init__(self, config: Config) -> None:
        """Initialize the evaluator with a Portia config and LLM scorer.

        Args:
            config (Config): Portia configuration with access to default model.

        """
        self.config = config
        self.scorer = LLMScorer(config)

    def process_plan(self, stream_item: PlanStreamItem) -> list[StreamMetric]:
        """Evaluate a Plan (not executed) using LLM-based scoring.

        Args:
            stream_item (PlanStreamItem): The Plan to evaluate.

        Returns:
            list[Metric]: A list of metrics scored by the LLM.

        """
        task_data = stream_item.plan.model_dump_json()
        metrics = self.scorer.score(
            task_data=[task_data],
            metrics_to_score=[
                MetricOnly(
                    name="correctness",
                    description="Are the steps logically sound and valid?",
                ),
                MetricOnly(
                    name="completeness",
                    description="Are all necessary steps included?",
                ),
                MetricOnly(
                    name="clearness",
                    description="Are the steps clearly explained?",
                ),
            ],
        )

        return [
            StreamMetric.from_stream_item(
                stream_item=stream_item,
                score=m.score,
                name=m.name,
                description=m.description,
                explanation=m.explanation,
            )
            for m in metrics
        ]

    def process_plan_run(self, stream_item: PlanRunStreamItem) -> list[StreamMetric]:
        """Evaluate a PlanRun (executed plan) using LLM-based scoring.

        Args:
            stream_item (PlanRunStreamItem): The linked plan + plan_run to process.

        Returns:
            list[Metric]: A list of performance metrics scored by the LLM.

        """
        task_data = f"""
         plan: {stream_item.plan.model_dump_json()}
         plan_run: {stream_item.plan_run.model_dump_json()}
        """
        metrics = self.scorer.score(
            task_data=[task_data],
            metrics_to_score=[
                MetricOnly(
                    name="success",
                    description="Did it accomplish the intended goal?",
                ),
                MetricOnly(
                    name="efficiency",
                    description="Were the steps necessary and minimal?",
                ),
            ],
        )

        return [
            StreamMetric.from_stream_item(
                stream_item=stream_item,
                score=m.score,
                name=m.name,
                description=m.description,
                explanation=m.explanation,
            )
            for m in metrics
        ]



"""Contains implementations of evals for SteelThread."""

from .backend import PortiaBackend
from .default_evaluator import DefaultEvaluator
from .eval_runner import EvalConfig, EvalRunner
from .evaluator import Evaluator, PlanRunMetadata
from .metrics import (
    EvalLogMetricBackend,
    EvalMetric,
    PortiaEvalMetricsBackend,
)
from .models import EvalTestCase, InputConfig, OutcomeAssertion
from .tags import EvalMetricTagger

__all__ = [
    "DefaultEvaluator",
    "EvalConfig",
    "EvalLogMetricBackend",
    "EvalMetric",
    "EvalMetricTagger",
    "EvalRunner",
    "EvalTestCase",
    "Evaluator",
    "InputConfig",
    "OutcomeAssertion",
    "PlanRunMetadata",
    "PortiaBackend",
    "PortiaEvalMetricsBackend",
]



"""Custom Portia instance."""

from portia import Clarification, Plan, PlanRun, Portia


class NoAuthPullPortia(Portia):
    """Override pull auth forward."""

    # to avoid OAuth clarifications which go via the Batch endpoint we return no clarifications here
    # This will defer to the individual tool readiness checks which we can deal with in the Tool.
    def _check_remaining_tool_readiness(
        self,
        plan: Plan,  # noqa: ARG002
        plan_run: PlanRun,  # noqa: ARG002
        start_index: int | None = None,  # noqa: ARG002
    ) -> list[Clarification]:
        return []



"""Backend for Portia evals."""

import httpx
from portia import Config
from portia.storage import PortiaCloudClient
from pydantic import BaseModel

from steelthread.evals.models import EvalTestCase


class PortiaBackend(BaseModel):
    """Client interface for interacting with the Portia API for evaluations.

    Provides methods to load test cases and mark them as processed.

    Attributes:
        config (Config): The Portia configuration containing API credentials and context.

    """

    config: Config

    def client(self) -> httpx.Client:
        """Create an HTTP client for interacting with the Portia API.

        Returns:
            httpx.Client: A configured HTTP client.

        """
        return PortiaCloudClient().get_client(self.config)

    def check_response(self, response: httpx.Response) -> None:
        """Validate the response from Portia API.

        Args:
            response (httpx.Response): The response from the Portia API to check.

        Raises:
            ValueError: If the response status code indicates an error.

        """
        if not response.is_success:
            error_str = str(response.content)
            raise ValueError(error_str)

    def load_evals(self, dataset_name: str, run_id: str) -> list[EvalTestCase]:
        """Load test cases from the Portia API with pagination."""
        client = self.client()
        url = f"/api/v0/evals/dataset-test-cases/?dataset_name={dataset_name}"
        test_cases = []

        while url:
            response = client.get(url)
            self.check_response(response)
            data = response.json()
            test_cases.extend(
                EvalTestCase(**tc, testcase=tc["id"], run=run_id) for tc in data.get("results", [])
            )
            url = data.get("next")

        return test_cases
